const { matchesState } = require('../../lib/index');

// @ts-check
const { Machine } = require('../../lib/index');
const { getSimplePathsAsArray } = require('../../lib/graph');
const puppeteer = require('puppeteer');
const assert = require('chai').assert;

const createInstagramMachine = page =>
  Machine(
    {
      initial: 'home',
      states: {
        home: {
          initial: 'loading',
          states: {
            loading: {
              on: {
                HOME_LOADED: 'loaded'
              }
            },
            loaded: {}
          },
          on: {
            LOG_IN_LINK: 'login'
          }
        },
        login: {
          initial: 'start',
          states: {
            start: {
              on: {
                ENTER_USERNAME: 'username'
              }
            },
            username: {
              on: {
                ENTER_PASSWORD: 'password'
              }
            },
            password: {}
          },
          on: {
            LOG_IN: [
              { target: 'suspicious', cond: 'isSuspicious' },
              { target: 'feed' }
            ]
          }
        },
        suspicious: {},
        feed: {
          initial: 'loading',
          states: {
            loading: {
              on: {
                FEED_LOADED: 'loaded'
              }
            },
            loaded: {}
          },
          on: {
            PROFILE_CLICK: 'profile'
          }
        },
        profile: {}
      }
    },
    {
      guards: {
        isSuspicious: () => false
      }
    }
  );

async function init() {
  const instagramMachine = createInstagramMachine();
  const paths = getSimplePathsAsArray(instagramMachine);
  const browser = await puppeteer.launch({ headless: false });
  const page = await browser.newPage();
  await page.goto('https://instagram.com');

  const eventMap = {
    HOME_LOADED: async page => {
      return true;
    },
    LOG_IN_LINK: async page => {
      const link = await page.$('[href="/accounts/login/"]');
      await link.click();
    },
    ENTER_USERNAME: async page => {
      await page.waitFor('[aria-label="Phone number, username, or email"]');
      await new Promise(res => {
        setTimeout(() => {
          res();
        }, 2000);
      });
      const input = await page.$(
        '[aria-label="Phone number, username, or email"]'
      );
      await input.type('davidkpiano');
    },
    ENTER_PASSWORD: async page => {
      const input = await page.$('[type="password"]');
      await input.type('asdf!');
    },
    LOG_IN: async page => {
      const logInButton = await page.$(
        '#react-root > section > main > div > article > div > div:nth-child(1) > div > form > span > button'
      );
      await logInButton.click();
    },
    FEED_LOADED: async page => {
      await page.waitFor('article header a');
    },
    PROFILE_CLICK: async page => {
      const button = await page.$('article header a');
      console.log(button);
      await button.click();
    }
  };

  const profilePath = paths.filter(path => path.state === 'profile')[0];
  const subPaths = profilePath.paths[profilePath.paths.length - 1];

  console.dir(subPaths, { depth: null });

  let currentState = instagramMachine.initialState;

  for (let i = 0; i < subPaths.length; i++) {
    const subPath = subPaths[i];

    // make sure currentstate is correct
    console.log(subPath.state, currentState.value);
    assert.ok(matchesState(subPath.state, currentState.value));

    console.log('STATE', subPath.state);
    console.log('EVENT', subPath.event);

    // wait 1 second
    await new Promise(res => setTimeout(res, 1000));

    // execute event
    const execEvent = eventMap[subPath.event];

    await execEvent(page);

    currentState = instagramMachine.transition(currentState, subPath.event);
  }
}

init();
