"use strict";

const Redux = require(`redux`);

const Promise = require(`bluebird`);

const _ = require(`lodash`);

const fs = require(`fs`);

const mitt = require(`mitt`);

const stringify = require(`json-stringify-safe`); // Create event emitter for actions


const emitter = mitt(); // Reducers

const reducers = require(`./reducers`);

const objectToMap = obj => {
  let map = new Map();
  Object.keys(obj).forEach(key => {
    map.set(key, obj[key]);
  });
  return map;
};

const mapToObject = map => {
  const obj = {};

  for (var _iterator = map, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
    var _ref;

    if (_isArray) {
      if (_i >= _iterator.length) break;
      _ref = _iterator[_i++];
    } else {
      _i = _iterator.next();
      if (_i.done) break;
      _ref = _i.value;
    }

    let _ref2 = _ref,
        key = _ref2[0],
        value = _ref2[1];
    obj[key] = value;
  }

  return obj;
}; // Read from cache the old node data.


let initialState = {};

try {
  const file = fs.readFileSync(`${process.cwd()}/.cache/redux-state.json`); // Apparently the file mocking in node-tracking-test.js
  // can override the file reading replacing the mocked string with
  // an already parsed object.

  if (Buffer.isBuffer(file) || typeof file === `string`) {
    initialState = JSON.parse(file);
  }

  if (initialState.staticQueryComponents) {
    initialState.staticQueryComponents = objectToMap(initialState.staticQueryComponents);
  }

  if (initialState.components) {
    initialState.components = objectToMap(initialState.components);
  }

  if (initialState.nodes) {
    initialState.nodes = objectToMap(initialState.nodes);
  }
} catch (e) {// ignore errors.
}

const store = Redux.createStore(Redux.combineReducers(Object.assign({}, reducers)), initialState, Redux.applyMiddleware(function multi({
  dispatch
}) {
  return next => action => Array.isArray(action) ? action.filter(Boolean).map(dispatch) : next(action);
})); // Persist state.

const saveState = state => {
  const pickedState = _.pick(state, [`nodes`, `status`, `componentDataDependencies`, `jsonDataPaths`, `components`, `staticQueryComponents`]);

  pickedState.staticQueryComponents = mapToObject(pickedState.staticQueryComponents);
  pickedState.components = mapToObject(pickedState.components);
  pickedState.nodes = mapToObject(pickedState.nodes);
  const stringified = stringify(pickedState, null, 2);
  fs.writeFile(`${process.cwd()}/.cache/redux-state.json`, stringified, () => {});
};

const saveStateDebounced = _.debounce(saveState, 1000);

store.subscribe(() => {
  const lastAction = store.getState().lastAction;
  emitter.emit(lastAction.type, lastAction);
}); // During development, once bootstrap is finished, persist state on changes.

let bootstrapFinished = false;

if (process.env.gatsby_executing_command === `develop`) {
  emitter.on(`BOOTSTRAP_FINISHED`, () => {
    bootstrapFinished = true;
    saveState(store.getState());
  });
  emitter.on(`*`, () => {
    if (bootstrapFinished) {
      saveStateDebounced(store.getState());
    }
  });
} // During builds, persist state once bootstrap has finished.


if (process.env.gatsby_executing_command === `build`) {
  emitter.on(`BOOTSTRAP_FINISHED`, () => {
    saveState(store.getState());
  });
}
/** Event emitter */


exports.emitter = emitter;
/** Redux store */

exports.store = store;
/**
 * Get all nodes from redux store.
 *
 * @returns {Array}
 */

exports.getNodes = () => {
  const nodes = store.getState().nodes;

  if (nodes) {
    return Array.from(nodes.values());
  } else {
    return [];
  }
};

const getNode = id => store.getState().nodes.get(id);
/** Get node by id from store.
 *
 * @param {string} id
 * @returns {Object}
 */


exports.getNode = getNode;
/**
 * Determine if node has changed.
 *
 * @param {string} id
 * @param {string} digest
 * @returns {boolean}
 */

exports.hasNodeChanged = (id, digest) => {
  const node = store.getState().nodes.get(id);

  if (!node) {
    return true;
  } else {
    return node.internal.contentDigest !== digest;
  }
};
/**
 * Get content for a node from the plugin that created it.
 *
 * @param {Object} node
 * @returns {promise}
 */


exports.loadNodeContent = node => {
  if (_.isString(node.internal.content)) {
    return Promise.resolve(node.internal.content);
  } else {
    return new Promise(resolve => {
      // Load plugin's loader function
      const plugin = store.getState().flattenedPlugins.find(plug => plug.name === node.internal.owner);

      const _require = require(plugin.resolve),
            loadNodeContent = _require.loadNodeContent;

      if (!loadNodeContent) {
        throw new Error(`Could not find function loadNodeContent for plugin ${plugin.name}`);
      }

      return loadNodeContent(node).then(content => {
        // TODO update node's content field here.
        resolve(content);
      });
    });
  }
};
/**
 * Get node and save path dependency.
 *
 * @param {string} id
 * @param {string} path
 * @returns {Object} node
 */


exports.getNodeAndSavePathDependency = (id, path) => {
  const _require2 = require(`./actions/add-page-dependency`),
        createPageDependency = _require2.createPageDependency;

  const node = getNode(id);
  createPageDependency({
    path,
    nodeId: id
  });
  return node;
}; // Start plugin runner which listens to the store
// and invokes Gatsby API based on actions.


require(`./plugin-runner`);
//# sourceMappingURL=index.js.map