var __assign =
  (this && this.__assign) ||
  function() {
    __assign =
      Object.assign ||
      function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
          s = arguments[i];
          for (var p in s)
            if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
      };
    return __assign.apply(this, arguments);
  };
import { SpecialTargets, ActionTypes } from './types';
import * as actionTypes from './actionTypes';
import { toEventObject } from './actions';
import { Machine } from './Machine';
import { StateNode } from './StateNode';
var SimulatedClock = /** @class */ /*#__PURE__*/ (function() {
  function SimulatedClock() {
    this.timeouts = new Map();
    this._now = 0;
    this._id = 0;
  }
  SimulatedClock.prototype.now = function() {
    return this._now;
  };
  SimulatedClock.prototype.getId = function() {
    return this._id++;
  };
  SimulatedClock.prototype.setTimeout = function(fn, timeout) {
    var id = this.getId();
    this.timeouts.set(id, {
      start: this.now(),
      timeout: timeout,
      fn: fn
    });
    return id;
  };
  SimulatedClock.prototype.clearTimeout = function(id) {
    this.timeouts.delete(id);
  };
  SimulatedClock.prototype.set = function(time) {
    if (this._now > time) {
      throw new Error('Unable to travel back in time');
    }
    this._now = time;
    this.flushTimeouts();
  };
  SimulatedClock.prototype.flushTimeouts = function() {
    var _this = this;
    this.timeouts.forEach(function(timeout, id) {
      if (_this.now() - timeout.start >= timeout.timeout) {
        timeout.fn.call(null);
        _this.timeouts.delete(id);
      }
    });
  };
  SimulatedClock.prototype.increment = function(ms) {
    this._now += ms;
    this.flushTimeouts();
  };
  return SimulatedClock;
})();
export { SimulatedClock };
// tslint:disable-next-line:max-classes-per-file
var Interpreter = /** @class */ /*#__PURE__*/ (function() {
  function Interpreter(machine, options) {
    if (options === void 0) {
      options = Interpreter.defaultOptions;
    }
    var _this = this;
    this.machine = machine;
    this.eventQueue = [];
    this.delayedEventsMap = {};
    this.activitiesMap = {};
    this.listeners = new Set();
    this.contextListeners = new Set();
    this.stopListeners = new Set();
    this.doneListeners = new Set();
    this.eventListeners = new Set();
    this.sendListeners = new Set();
    this.initialized = false;
    this.children = new Set();
    this.init = this.start;
    this.send = function(event) {
      var eventObject = toEventObject(event);
      if (!_this.initialized) {
        throw new Error(
          'Unable to send event "' +
            eventObject.type +
            '" to an uninitialized interpreter (ID: ' +
            _this.machine.id +
            '). Event: ' +
            JSON.stringify(event)
        );
      }
      var nextState = _this.machine.transition(
        _this.state,
        eventObject,
        _this.state.context
      ); // TODO: fixme
      _this.update(nextState, event);
      _this.flushEventQueue();
      // Forward copy of event to child interpreters
      _this.forward(eventObject);
      return nextState;
      // tslint:disable-next-line:semicolon
    };
    var resolvedOptions = __assign({}, Interpreter.defaultOptions, options);
    this.clock = resolvedOptions.clock;
    this.logger = resolvedOptions.logger;
    this.parent = resolvedOptions.parent;
  }
  Object.defineProperty(Interpreter.prototype, 'initialState', {
    /**
     * The initial state of the statechart.
     */
    get: function() {
      return this.machine.initialState;
    },
    enumerable: true,
    configurable: true
  });
  Interpreter.prototype.update = function(state, event) {
    var _this = this;
    this.state = state;
    var context = this.state.context;
    var eventObject = event ? toEventObject(event) : undefined;
    this.state.actions.forEach(function(action) {
      _this.exec(action, context, eventObject);
    }, context);
    if (eventObject) {
      this.eventListeners.forEach(function(listener) {
        return listener(eventObject);
      });
    }
    this.listeners.forEach(function(listener) {
      return listener(state);
    });
    this.contextListeners.forEach(function(ctxListener) {
      return ctxListener(
        _this.state.context,
        _this.state.history ? _this.state.history.context : undefined
      );
    });
    if (this.state.tree && this.state.tree.done) {
      this.doneListeners.forEach(function(listener) {
        return listener(state);
      });
      this.stop();
    }
  };
  /*
     * Adds a listener that is called whenever a state transition happens.
     * @param listener The listener to add
     */
  Interpreter.prototype.onTransition = function(listener) {
    this.listeners.add(listener);
    return this;
  };
  Interpreter.prototype.onEvent = function(listener) {
    this.eventListeners.add(listener);
    return this;
  };
  Interpreter.prototype.onSend = function(listener) {
    this.sendListeners.add(listener);
    return this;
  };
  Interpreter.prototype.onChange = function(listener) {
    this.contextListeners.add(listener);
    return this;
  };
  Interpreter.prototype.onStop = function(listener) {
    this.stopListeners.add(listener);
    return this;
  };
  Interpreter.prototype.onDone = function(listener) {
    this.doneListeners.add(listener);
    return this;
  };
  /**
   * Removes a listener.
   * @param listener The listener to remove
   */
  Interpreter.prototype.off = function(listener) {
    this.listeners.delete(listener);
    return this;
  };
  Interpreter.prototype.start = function(initialState) {
    if (initialState === void 0) {
      initialState = this.machine.initialState;
    }
    this.initialized = true;
    this.update(initialState);
    return this;
  };
  Interpreter.prototype.stop = function() {
    var _this = this;
    this.listeners.forEach(function(listener) {
      return _this.off(listener);
    });
    this.stopListeners.forEach(function(listener) {
      // call listener, then remove
      listener();
      _this.stopListeners.delete(listener);
    });
    this.contextListeners.forEach(function(ctxListener) {
      return _this.contextListeners.delete(ctxListener);
    });
    this.doneListeners.forEach(function(doneListener) {
      return _this.doneListeners.delete(doneListener);
    });
    return this;
  };
  Interpreter.prototype.forward = function(event) {
    this.children.forEach(function(childInterpreter) {
      return childInterpreter.send(event);
    });
  };
  Interpreter.prototype.defer = function(sendAction) {
    var _this = this;
    return this.clock.setTimeout(function() {
      return _this.send(sendAction.event);
    }, sendAction.delay || 0);
  };
  Interpreter.prototype.cancel = function(sendId) {
    this.clock.clearTimeout(this.delayedEventsMap[sendId]);
    delete this.delayedEventsMap[sendId];
  };
  Interpreter.prototype.exec = function(action, context, event) {
    var _this = this;
    if (action.exec) {
      return action.exec(context, event);
    }
    switch (action.type) {
      case actionTypes.send:
        var sendAction = action;
        switch (sendAction.to) {
          case SpecialTargets.Parent:
            if (this.parent) {
              this.parent.send(sendAction.event);
            }
            break;
          default:
            if (!sendAction.delay) {
              this.eventQueue.push(sendAction.event);
            } else {
              this.delayedEventsMap[sendAction.id] = this.defer(sendAction);
            }
            break;
        }
      case actionTypes.cancel:
        this.cancel(action.sendId);
        break;
      case actionTypes.start: {
        var activity = action.activity;
        if (activity.type === ActionTypes.Invoke) {
          var service =
            this.machine.options.services && activity.src
              ? this.machine.options.services[activity.src]
              : undefined;
          var autoForward = !!activity.forward;
          if (!service) {
            console.warn(
              "No service found for invocation '" + activity.src + "'"
            );
            return;
          }
          if (typeof service !== 'string') {
            // TODO: try/catch here
            var childMachine =
              service instanceof StateNode ? service : Machine(service);
            var interpreter_1 = this.spawn(childMachine, autoForward);
            interpreter_1.start();
            this.activitiesMap[activity.id] = function() {
              _this.children.delete(interpreter_1);
              interpreter_1.stop();
            };
          }
        } else {
          var implementation =
            this.machine.options && this.machine.options.activities
              ? this.machine.options.activities[activity.type]
              : undefined;
          if (!implementation) {
            console.warn(
              "No implementation found for activity '" + activity.type + "'"
            );
            return;
          }
          // Start implementation
          this.activitiesMap[activity.id] = implementation(context, activity);
        }
        break;
      }
      case actionTypes.stop: {
        var activity = action.activity;
        var dispose = this.activitiesMap[activity.id];
        if (dispose) {
          dispose();
        }
        break;
      }
      case actionTypes.log:
        var expr = action.expr ? action.expr(context, event) : undefined;
        if (action.label) {
          this.logger(action.label, expr);
        } else {
          this.logger(expr);
        }
        break;
      default:
        // tslint:disable-next-line:no-console
        console.warn(
          "No implementation found for action type '" + action.type + "'"
        );
        break;
    }
    return undefined;
  };
  Interpreter.prototype.spawn = function(machine, autoForward) {
    if (autoForward === void 0) {
      autoForward = false;
    }
    var childInterpreter = new Interpreter(machine, {
      parent: this
    });
    if (autoForward) {
      this.children.add(childInterpreter);
    }
    return childInterpreter;
  };
  Interpreter.prototype.flushEventQueue = function() {
    var flushedEvent = this.eventQueue.shift();
    if (flushedEvent) {
      this.send(flushedEvent);
    }
  };
  // TODO: fixme
  Interpreter.defaultOptions = {
    clock: {
      setTimeout: function(fn, ms) {
        return window.setTimeout.call(null, fn, ms);
      },
      clearTimeout: function(id) {
        return window.clearTimeout.call(null, id);
      }
    },
    logger: /*#__PURE__*/ global.console.log.bind(console)
  };
  Interpreter.interpret = interpret;
  return Interpreter;
})();
export { Interpreter };
export function interpret(machine, options) {
  var interpreter = new Interpreter(machine, options);
  return interpreter;
}
