"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
var constants_1 = require("./constants");
var utils_1 = require("./utils");
function stateValuesEqual(a, b) {
    if (a === b) {
        return true;
    }
    var aKeys = utils_1.keys(a);
    var bKeys = utils_1.keys(b);
    return (aKeys.length === bKeys.length &&
        aKeys.every(function (key) { return stateValuesEqual(a[key], b[key]); }));
}
exports.stateValuesEqual = stateValuesEqual;
var State = /** @class */ (function () {
    /**
     * Creates a new State instance.
     * @param value The state value
     * @param context The extended state
     * @param historyValue The tree representing historical values of the state nodes
     * @param history The previous state
     * @param actions An array of action objects to execute as side-effects
     * @param activities A mapping of activities and whether they are started (`true`) or stopped (`false`).
     * @param data
     * @param events Internal event queue. Should be empty with run-to-completion semantics.
     * @param tree
     */
    function State(value, context, historyValue, history, actions, activities, data, events, tree) {
        if (actions === void 0) { actions = []; }
        if (activities === void 0) { activities = constants_1.EMPTY_ACTIVITY_MAP; }
        if (data === void 0) { data = {}; }
        if (events === void 0) { events = []; }
        this.value = value;
        this.context = context;
        this.historyValue = historyValue;
        this.history = history;
        this.actions = actions;
        this.activities = activities;
        this.data = data;
        this.events = events;
        Object.defineProperty(this, 'tree', {
            value: tree,
            enumerable: false
        });
    }
    /**
     * Creates a new State instance for the given `stateValue` and `context`.
     * @param stateValue
     * @param context
     */
    State.from = function (stateValue, context) {
        if (stateValue instanceof State) {
            if (stateValue.context !== context) {
                return new State(stateValue.value, context, stateValue.historyValue, stateValue.history, [], stateValue.activities, {}, [], stateValue.tree);
            }
            return stateValue;
        }
        return new State(stateValue, context, undefined, undefined, [], undefined, undefined, []);
    };
    /**
     * Creates a new State instance for the given `stateValue` and `context` with no actions (side-effects).
     * @param stateValue
     * @param context
     */
    State.inert = function (stateValue, context) {
        if (stateValue instanceof State) {
            if (!stateValue.actions.length) {
                return stateValue;
            }
            return new State(stateValue.value, context, stateValue.historyValue, stateValue.history, undefined, stateValue.activities, undefined, undefined, stateValue.tree);
        }
        return State.from(stateValue, context);
    };
    Object.defineProperty(State.prototype, "nextEvents", {
        /**
         * The next events that will cause a transition from the current state.
         */
        get: function () {
            if (!this.tree) {
                return [];
            }
            return this.tree.nextEvents;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Returns an array of all the string leaf state node paths.
     * @param stateValue
     * @param delimiter The character(s) that separate each subpath in the string state node path.
     */
    State.prototype.toStrings = function (stateValue, delimiter) {
        var _this = this;
        if (stateValue === void 0) { stateValue = this.value; }
        if (delimiter === void 0) { delimiter = '.'; }
        if (typeof stateValue === 'string') {
            return [stateValue];
        }
        var valueKeys = utils_1.keys(stateValue);
        return valueKeys.concat.apply(valueKeys, __spread(valueKeys.map(function (key) {
            return _this.toStrings(stateValue[key]).map(function (s) { return key + delimiter + s; });
        })));
    };
    /**
     * Whether the current state value is a subset of the given parent state value.
     * @param parentStateValue
     */
    State.prototype.matches = function (parentStateValue) {
        return utils_1.matchesState(parentStateValue, this.value);
    };
    Object.defineProperty(State.prototype, "changed", {
        /**
         * Indicates whether the state has changed from the previous state. A state is considered "changed" if:
         *
         * - Its value is not equal to its previous value, or:
         * - It has any new actions (side-effects) to execute.
         *
         * An initial state (with no history) will return `undefined`.
         */
        get: function () {
            if (!this.history) {
                return undefined;
            }
            return (!!this.actions.length ||
                (typeof this.history.value !== typeof this.value
                    ? true
                    : typeof this.value === 'string'
                        ? this.value !== this.history.value
                        : stateValuesEqual(this.value, this.history.value)));
        },
        enumerable: true,
        configurable: true
    });
    return State;
}());
exports.State = State;
